"""vendored code copied from black

For the license, see /licenses/black
"""

import os
import sys
from functools import lru_cache
from pathlib import Path

import tomli
from pathspec import PathSpec


@lru_cache
def find_project_root(srcs):
    """Return a directory containing .git, .hg, or pyproject.toml.

    That directory will be a common parent of all files and directories
    passed in `srcs`.

    If no directory in the tree contains a marker that would specify it's the
    project root, the root of the file system is returned.

    Returns a two-tuple with the first element as the project root path and
    the second element as a string describing the method by which the
    project root was discovered.
    """
    if not srcs:
        srcs = [str(Path.cwd().resolve())]

    path_srcs = [Path(Path.cwd(), src).resolve() for src in srcs]

    # A list of lists of parents for each 'src'. 'src' is included as a
    # "parent" of itself if it is a directory
    src_parents = [
        list(path.parents) + ([path] if path.is_dir() else []) for path in path_srcs
    ]

    common_base = max(
        set.intersection(*(set(parents) for parents in src_parents)),
        key=lambda path: path.parts,
    )

    for directory in (common_base, *common_base.parents):
        if (directory / ".git").exists():
            return directory, ".git directory"

        if (directory / ".hg").is_dir():
            return directory, ".hgdirectory"

        if (directory / "pyproject.toml").is_file():
            return directory, "pyproject.toml"

    return directory, "file system root"


def find_pyproject_toml(path_search_start):
    """Find the absolute filepath to a pyproject.toml if it exists"""
    path_project_root, _ = find_project_root(path_search_start)
    path_pyproject_toml = path_project_root / "pyproject.toml"
    if path_pyproject_toml.is_file():
        return str(path_pyproject_toml)

    try:
        path_user_pyproject_toml = find_user_pyproject_toml()
        return (
            str(path_user_pyproject_toml)
            if path_user_pyproject_toml.is_file()
            else None
        )
    except (PermissionError, RuntimeError) as e:
        # We do not have access to the user-level config directory, so ignore it.
        print(f"Ignoring user configuration directory due to {e!r}")
        return None


@lru_cache
def find_user_pyproject_toml():
    r"""Return the path to the top-level user configuration for black.

    This looks for ~\.black on Windows and ~/.config/black on Linux and other
    Unix systems.

    May raise:
    - RuntimeError: if the current user has no homedir
    - PermissionError: if the current process cannot access the user's homedir
    """
    if sys.platform == "win32":
        # Windows
        user_config_path = Path.home() / ".black"
    else:
        config_root = os.environ.get("XDG_CONFIG_HOME", "~/.config")
        user_config_path = Path(config_root).expanduser() / "black"
    return user_config_path.resolve()


def parse_pyproject_toml(path_config):
    """Parse a pyproject toml file, pulling out relevant parts for Black and Blackdoc

    If parsing fails, will raise a tomli.TOMLDecodeError
    """
    with open(path_config, "rb") as f:
        pyproject_toml = tomli.load(f)

    black_config = pyproject_toml.get("tool", {}).get("black", {})
    blackdoc_config = pyproject_toml.get("tool", {}).get("blackdoc", {})
    config = {**black_config, **blackdoc_config}

    return {k.replace("--", "").replace("-", "_"): v for k, v in config.items()}


def read_pyproject_toml(source, config_path):
    if not config_path:
        config_path = find_pyproject_toml(source)
        if config_path is None:
            return {}

    try:
        config = parse_pyproject_toml(config_path)
    except (ValueError, OSError) as e:
        raise OSError(
            f"Error reading configuration file ({config_path}): {e}"
        ) from None

    if not config:
        return {}

    return config


def normalize_path_maybe_ignore(path, root, report):
    """Normalize `path`. May return `None` if `path` was ignored.

    `report` is where "path ignored" output goes.
    """
    try:
        abspath = path if path.is_absolute() else Path.cwd() / path
        normalized_path = abspath.resolve().relative_to(root).as_posix()
    except OSError as e:
        report.path_ignored(path, f"cannot be read because {e}")
        return None

    except ValueError:
        if path.is_symlink():
            report.path_ignored(path, f"is a symbolic link that points outside {root}")
            return None

        raise

    return normalized_path


def path_is_excluded(normalized_path, pattern):
    match = pattern.search(normalized_path) if pattern else None
    return bool(match and match.group(0))


@lru_cache
def get_gitignore(root):
    """Return a PathSpec matching gitignore content if present."""
    gitignore = root / ".gitignore"
    lines = []
    if gitignore.is_file():
        with gitignore.open(encoding="utf-8") as gf:
            lines = gf.readlines()
    return PathSpec.from_lines("gitwildmatch", lines)


@lru_cache
def jupyter_dependencies_are_installed(*, verbose, quiet):
    try:
        import IPython  # noqa:F401
        import tokenize_rt  # noqa:F401
    except ModuleNotFoundError:
        if verbose or not quiet:
            msg = (
                "Skipping .ipynb files as Jupyter dependencies are not installed.\n"
                "You can fix this by running ``pip install black[jupyter]``"
            )
            print(msg)
        return False
    else:
        return True


def gen_python_files(
    paths,
    root,
    include,
    exclude,
    extend_exclude,
    force_exclude,
    report,
    gitignore,
    *,
    verbose,
    quiet,
):
    """Generate all files under `path` whose paths are not excluded by the
    `exclude_regex` or `force_exclude` regexes, but are included by the `include` regex.

    Symbolic links pointing outside of the `root` directory are ignored.

    `report` is where output about exclusions goes.
    """
    assert root.is_absolute(), f"INTERNAL ERROR: `root` must be absolute but is {root}"
    for child in paths:
        normalized_path = normalize_path_maybe_ignore(child, root, report)
        if normalized_path is None:
            continue

        # First ignore files matching .gitignore, if passed
        if gitignore is not None and gitignore.match_file(normalized_path):
            report.path_ignored(child, "matches the .gitignore file content")
            continue

        # Then ignore with `--exclude` `--exent-exclude` and `--force-exclude` options.
        normalized_path = "/" + normalized_path
        if child.is_dir():
            normalized_path += "/"

        if path_is_excluded(normalized_path, exclude):
            report.path_ignored(child, "matches the --exclude regular expression")
            continue

        if path_is_excluded(normalized_path, extend_exclude):
            report.path_ignored(
                child, "matches the --extend-exclude regular expression"
            )
            continue

        if path_is_excluded(normalized_path, force_exclude):
            report.path_ignored(child, "matches the --force-exclude regular expression")
            continue

        if child.is_dir():
            # If gitignore is None, gitignore usage is disabled, while a Falsey
            # gitignore is when the directory doesn't have a .gitignore file.
            yield from gen_python_files(
                child.iterdir(),
                root,
                include,
                exclude,
                extend_exclude,
                force_exclude,
                report,
                gitignore + get_gitignore(child) if gitignore is not None else None,
                verbose=verbose,
                quiet=quiet,
            )

        elif child.is_file():
            # if child.suffix == ".ipynb" and not jupyter_dependencies_are_installed(
            #     verbose=verbose, quiet=quiet
            # ):
            #     continue
            include_match = include.search(normalized_path) if include else True
            if include_match:
                yield child



import itertools
import re

import more_itertools

from blackdoc.formats.errors import InvalidFormatError

name = "ipython"

prompt_re = re.compile(r"^(?P<indent>[ ]*)(?P<prompt>In \[(?P<count>\d+)\]: )")
continuation_prompt_re = re.compile(r"^(?P<indent>[ ]*)\.\.\.: ")

prompt_template = "In [{count}]: "
continuation_template = "...: "

magic_re = re.compile(r"^(!.*|%.*|@[a-zA-Z_][a-zA-Z0-9_]* .+)")
magic_comment = "<ipython-magic>"

include_pattern = r"\.pyi?$"


def continuation_lines(lines, indent, prompt_length):
    while True:
        try:
            line_number, line = lines.peek()
        except StopIteration:
            line_number = -1
            line = ""

        match = continuation_prompt_re.match(line)
        if not match or len(match.groupdict()["indent"]) - prompt_length + 5 != indent:
            break

        # actually consume the item
        more_itertools.consume(lines, n=1)

        yield line_number, line


def detection_func(lines):
    try:
        _, line = lines.peek()
    except StopIteration:
        line = ""

    match = prompt_re.match(line)
    if not match:
        return None

    groups = match.groupdict()
    indent = len(groups["indent"])
    prompt_length = len(groups["prompt"])

    detected_lines = list(
        itertools.chain(
            [more_itertools.first(lines)],
            continuation_lines(lines, indent, prompt_length),
        )
    )
    line_numbers, lines = map(tuple, more_itertools.unzip(detected_lines))

    line_range = min(line_numbers), max(line_numbers) + 1
    if line_numbers != tuple(range(line_range[0], line_range[1])):
        raise InvalidFormatError("line numbers are not contiguous")

    return line_range, name, "\n".join(lines)


def is_ipython(line):
    is_prompt = prompt_re.match(line)
    is_continuation_prompt = continuation_prompt_re.match(line)
    return is_prompt or is_continuation_prompt


def metadata(line):
    match = prompt_re.match(line)
    if not match:
        return {}

    groups = match.groupdict()
    return {"count": int(groups["count"])}


def hide_magic(code):
    def comment_magic(line):
        stripped = line.lstrip()
        indent = len(line) - len(stripped)

        if not stripped or not magic_re.match(stripped):
            return line

        return " " * indent + f"# {magic_comment}" + stripped

    lines = code.split("\n")
    processed = tuple(comment_magic(line) for line in lines)

    return "\n".join(processed)


def reveal_magic(code):
    def uncomment_magic(line):
        stripped = line.lstrip()

        if magic_comment not in line:
            return line

        indent = len(line) - len(stripped)
        return " " * indent + stripped[len(magic_comment) + 2 :]

    lines = code.split("\n")
    processed = tuple(uncomment_magic(line) for line in lines)

    return "\n".join(processed)


def extraction_func(code):
    def remove_prompt(line, count):
        n = len(prompt_template.format(count=count))
        return line[n:]

    lines = code.split("\n")
    if len(lines) == 0:
        raise InvalidFormatError("no lines found")

    parameters = metadata(lines[0])

    if not all(is_ipython(line) for line in lines):
        raise InvalidFormatError(f"misformatted code unit: {code}")

    extracted = "\n".join(remove_prompt(line, **parameters) for line in lines)

    return parameters, hide_magic(extracted)


def reformatting_func(line, count):
    prompt = prompt_template.format(count=count)
    continuation_prompt = (
        " " * (len(prompt) - len(continuation_template)) + continuation_template
    )

    lines = iter(reveal_magic(line).split("\n"))

    reformatted = "\n".join(
        itertools.chain(
            more_itertools.always_iterable(prompt + more_itertools.first(lines)),
            (continuation_prompt + line for line in lines),
        )
    )

    return reformatted



import argparse
import pathlib
import sys

import black
from rich.text import Text

from blackdoc import __version__, format_lines, formats
from blackdoc.blackcompat import read_pyproject_toml
from blackdoc.colors import DiffHighlighter
from blackdoc.console import err, out
from blackdoc.diff import unified_diff
from blackdoc.files import collect_files
from blackdoc.report import Report

diff_highlighter = DiffHighlighter()


def check_format_names(string):
    names = string.split(",")
    allowed_names = set(formats.detection_funcs.keys()) - set(["none"])
    for name in names:
        if name in allowed_names:
            continue

        raise argparse.ArgumentTypeError(
            f"invalid choice: {name!r} (choose from {', '.join(sorted(allowed_names))})"
        )
    return names


def format_and_overwrite(path, mode):
    try:
        with open(path, mode="rb") as f:
            content, encoding, newline = black.decode_bytes(f.read())

        lines = content.split("\n")

        new_content = "\n".join(format_lines(lines, mode))

        if new_content == content:
            result = "unchanged"
        else:
            err.print(f"reformatted {path}", style="bold", highlight=False)
            result = "reformatted"

            with open(path, "w", encoding=encoding, newline=newline) as f:
                f.write(new_content)
    except (black.InvalidInput, formats.InvalidFormatError) as e:
        err.print(
            f"error: cannot format {path.absolute()}: {e}", style="red", highlight=False
        )
        result = "error"

    return result


def format_and_check(path, mode, diff=False, color=False):
    try:
        with open(path, mode="rb") as f:
            content, _, _ = black.decode_bytes(f.read())

        lines = content.split("\n")

        new_content = "\n".join(format_lines(lines, mode))

        if new_content == content:
            result = "unchanged"
        else:
            err.print(f"would reformat {path}", style="bold", highlight=False)

            if diff:
                diff_ = unified_diff(content, new_content, path)

                if color:
                    formatted_diff = diff_highlighter(diff_)
                else:
                    formatted_diff = Text(diff_)

                out.print(formatted_diff)

            result = "reformatted"
    except (black.InvalidInput, formats.InvalidFormatError) as e:
        err.print(
            f"error: cannot format {path.absolute()}: {e}", style="red", highlight=False
        )
        result = "error"

    return result


def process(args):
    if not args.src:
        err.print("No Path provided. Nothing to do :sleeping:", style="bold")
        return 0

    selected_formats = getattr(args, "formats", None)
    if selected_formats:
        formats.disable(
            set(formats.detection_funcs.keys()) - set(selected_formats) - {"none"}
        )

    disabled_formats = getattr(args, "disable_formats", None)
    if disabled_formats:
        formats.disable(disabled_formats)

    try:
        include_regex = black.re_compile_maybe_verbose(args.include)
    except black.re.error:
        err.print(
            f"Invalid regular expression for include given: {args.include!r}",
            style="red",
        )
        return 2

    try:
        exclude_regex = black.re_compile_maybe_verbose(args.exclude)
    except black.re.error:
        err.print(
            f"Invalid regular expression for exclude given: {args.exclude!r}",
            style="red",
        )
        return 2

    try:
        extend_exclude_regex = black.re_compile_maybe_verbose(args.extend_exclude)
    except black.re.error:
        err.print(
            f"Invalid regular expression for extend exclude given: {args.extend_exclude!r}",
            style="red",
        )
        return 2

    try:
        force_exclude = getattr(args, "force_exclude", "")
        force_exclude_regex = (
            black.re_compile_maybe_verbose(force_exclude) if force_exclude else None
        )
    except black.re.error:
        err.print(
            f"Invalid regular expression for force_exclude given: {force_exclude!r}",
            style="red",
        )
        return 2

    sources = set(
        collect_files(
            args.src,
            include_regex,
            exclude_regex,
            extend_exclude_regex,
            force_exclude_regex,
            quiet=args.quiet,
            verbose=args.verbose,
        )
    )
    if len(sources) == 0:
        err.print(
            "No files are present to be formatted. Nothing to do :sleeping:",
            style="bold",
        )
        return 0

    target_versions = set(
        black.TargetVersion[version.upper()]
        for version in getattr(args, "target_versions", ())
    )
    mode = black.Mode(
        line_length=args.line_length,
        target_versions=target_versions,
        string_normalization=not args.skip_string_normalization,
    )

    actions = {
        "inplace": format_and_overwrite,
        "check": format_and_check,
    }
    action_kwargs = {"diff": args.diff, "color": args.color} if args.diff else {}

    action = actions.get(args.action)

    changed_sources = {
        source: action(source.resolve(), mode, **action_kwargs)
        for source in sorted(sources)
    }

    conditional = args.action == "check"
    report = Report.from_sources(changed_sources, conditional=conditional)

    if report.n_error > 0:
        return_code = 123
    elif args.action == "check" and report.n_reformatted > 0:
        return_code = 1
    else:
        return_code = 0

    error_message = "Oh no! :boom: :broken_heart: :boom:"
    no_error_message = "All done! :sparkles: :cake: :sparkles:"
    err.print()
    err.print(
        error_message if report.n_error > 0 else no_error_message,
        style="bold",
    )
    err.print(report, highlight=False)
    return return_code


class boolean_flag(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        super().__init__(option_strings, dest, nargs=0, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        value = False if option_string.startswith("--no-") else True
        setattr(namespace, self.dest, value)


def main():
    program = pathlib.Path(__file__).parent.name

    parser = argparse.ArgumentParser(
        description="run black on documentation code snippets (e.g. doctest)",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        prog=program,
    )
    parser.add_argument(
        "-t",
        "--target-versions",
        action="append",
        choices=[v.name.lower() for v in black.TargetVersion],
        help=(
            "Python versions that should be supported by Black's output."
            " (default: per-file auto-detection)"
        ),
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "-l",
        "--line-length",
        metavar="INT",
        type=int,
        default=black.DEFAULT_LINE_LENGTH,
        help="How many characters per line to allow.",
    )
    parser.add_argument(
        "--check",
        dest="action",
        action="store_const",
        const="check",
        default="inplace",
        help=(
            "Don't write the files back, just return the status.  Return code 0"
            " means nothing would change.  Return code 1 means some files would be"
            " reformatted.  Return code 123 means there was an internal error."
        ),
    )
    parser.add_argument(
        "--diff",
        dest="diff",
        action="store_const",
        const="diff",
        help="Don't write the files back, just output a diff for each file on stdout.",
    )
    parser.add_argument(
        "--color",
        "--no-color",
        dest="color",
        action=boolean_flag,
        default=False,
        help="Show colored diff. Only applies when `--diff` is given.",
    )
    parser.add_argument(
        "--include",
        metavar="TEXT",
        type=str,
        default=formats.format_include_patterns(),
        help=(
            "A regular expression that matches files and directories that should be"
            " included on recursive searches.  An empty value means all files are"
            " included regardless of the name.  Use forward slashes for directories on"
            " all platforms (Windows, too).  Exclusions are calculated first, inclusions"
            " later."
        ),
    )
    parser.add_argument(
        "--exclude",
        metavar="TEXT",
        type=str,
        default=black.DEFAULT_EXCLUDES,
        help=(
            "A regular expression that matches files and directories that should be"
            " excluded on recursive searches.  An empty value means no paths are excluded."
            " Use forward slashes for directories on all platforms (Windows, too)."
            "  Exclusions are calculated first, inclusions later."
        ),
    )
    parser.add_argument(
        "--extend-exclude",
        metavar="TEXT",
        type=str,
        default="",
        help=(
            "Like --exclude, but adds additional files and directories"
            " on top of the excluded ones. (Useful if you simply want to"
            " add to the default)"
        ),
    )
    parser.add_argument(
        "--force-exclude",
        metavar="TEXT",
        type=str,
        default=argparse.SUPPRESS,
        help=(
            "Like --exclude, but files and directories"
            " matching this regex will be excluded even"
            " when they are passed explicitly as arguments"
        ),
    )
    parser.add_argument(
        "--formats",
        metavar="FMT[,FMT[,FMT...]]",
        type=check_format_names,
        help="Use only the specified formats.",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--disable-formats",
        metavar="FMT[,FMT[,FMT...]]",
        type=check_format_names,
        help=(
            "Disable the given formats."
            " This option also affects formats explicitly set."
        ),
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "-S",
        "--skip-string-normalization",
        dest="skip_string_normalization",
        action="store_true",
        help="Don't normalize string quotes or prefixes.",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help=(
            "Don't emit non-error messages to stderr. Errors are still"
            " emitted; silence those with 2>/dev/null."
        ),
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help=(
            "Also emit messages to stderr about files that were not"
            " changed or were ignored due to exclusion patterns."
        ),
    )
    parser.add_argument(
        "--version",
        action="version",
        help="Show the version and exit.",
        version=f"{program} {__version__}",
    )
    parser.add_argument(
        "--config",
        action="store",
        nargs=1,
        type=pathlib.Path,
        default=None,
        help="Read configuration from FILE path.",
    )
    parser.add_argument(
        "src",
        action="store",
        type=pathlib.Path,
        nargs="*",
        default=None,
        help="one or more paths to work on",
    )
    args = parser.parse_args()
    if args.config or args.src:
        file_defaults = read_pyproject_toml(tuple(args.src), args.config)
        parser.set_defaults(**file_defaults)

    if args.diff:
        parser.set_defaults(action="check")

    args = parser.parse_args()
    sys.exit(process(args))


if __name__ == "__main__":
    main()



import textwrap

import more_itertools
import pytest

from blackdoc import blacken
from blackdoc.formats import rst
from blackdoc.tests.data import rst as data


@pytest.mark.parametrize(
    ("string", "expected"),
    (
        pytest.param("", None, id="empty string"),
        pytest.param("Some string.", None, id="no_code"),
        pytest.param(
            textwrap.dedent(
                """\
                .. note::

                    This is not a code block.
                """
            ),
            None,
            id="block",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. code:: sh

                    find . -name "*.py"
                """
            ),
            None,
            id="code_other_language",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. code:: python

                    10 * 5
                """
            ),
            "rst",
            id="code",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. code-block:: python

                    10 * 5
                """
            ),
            "rst",
            id="code-block",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. ipython:: python

                    %%time
                    10 * 5
                """
            ),
            "rst",
            id="ipython",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. ipython::

                    In [1]: 10 * 5
                    Out[1]: 50

                    In [2]: %%time
                       ...: ".".join("abc")
                    Out[2]: 'a.b.c'
                """
            ),
            None,
            id="ipython-prompt",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. ipython::
                    :okerror:

                    @verbatim
                    In [1]: 10 * 5
                    Out[1]: 50
                """
            ),
            None,
            id="ipython-prompt-cell-decorator",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. testsetup::

                    10 * 5
                """
            ),
            "rst",
            id="testsetup",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. testcode::

                    10 * 5
                """
            ),
            "rst",
            id="testcode",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. testcleanup::

                    10 * 5
                """
            ),
            "rst",
            id="testcleanup",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. ipython:: python
                    print("abc")
                """
            ),
            "rst",
            id="missing option separator",
        ),
    ),
)
def test_detection_func(string, expected):
    def construct_expected(label, string):
        if label is None:
            return None

        n_lines = len(string.split("\n"))

        range_ = (1, n_lines + 1)
        return range_, label, string

    lines = string.split("\n")
    code_fragment = more_itertools.peekable(enumerate(lines, start=1))
    actual = rst.detection_func(code_fragment)

    assert actual == construct_expected(expected, string.rstrip())


@pytest.mark.parametrize(
    "code,expected",
    (
        pytest.param(
            textwrap.dedent(
                """\
                .. code:: python

                   10 * 5
                """.rstrip()
            ),
            (
                {
                    "name": "code",
                    "language": "python",
                    "options": (),
                    "prompt_length": 3,
                    "n_header_lines": 2,
                },
                "10 * 5",
            ),
            id="code",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. code:: python
                   :okwarning:

                   10 * 5
                """.rstrip()
            ),
            (
                {
                    "name": "code",
                    "language": "python",
                    "options": (":okwarning:",),
                    "prompt_length": 3,
                    "n_header_lines": 3,
                },
                "10 * 5",
            ),
            id="code_with_options",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. code-block:: python

                   10 * 5
                """.rstrip()
            ),
            (
                {
                    "name": "code-block",
                    "language": "python",
                    "options": (),
                    "prompt_length": 3,
                    "n_header_lines": 2,
                },
                "10 * 5",
            ),
            id="code-block",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. ipython::

                    %%time
                    10 * 5
                """.rstrip()
            ),
            (
                {
                    "name": "ipython",
                    "language": None,
                    "options": (),
                    "prompt_length": 4,
                    "n_header_lines": 2,
                },
                textwrap.dedent(
                    """\
                    # <ipython-magic>%%time
                    10 * 5
                    """.rstrip()
                ),
            ),
            id="ipython",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. ipython:: python
                    10 * 5
                """.rstrip()
            ),
            (
                {
                    "name": "ipython",
                    "language": "python",
                    "options": (),
                    "prompt_length": 4,
                    "n_header_lines": 2,
                },
                "10 * 5",
            ),
            id="missing_eof_line",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                .. ipython:: python
                    print("abc")
                """.rstrip()
            ),
            (
                {
                    "name": "ipython",
                    "language": "python",
                    "options": (),
                    "prompt_length": 4,
                    "n_header_lines": 2,
                },
                'print("abc")',
            ),
            id="missing_sep_line",
        ),
    ),
)
def test_extraction_func(code, expected):
    actual = rst.extraction_func(code)

    assert expected == actual


@pytest.mark.parametrize(
    ("code", "directive", "expected"),
    (
        pytest.param(
            "10 * 5",
            {
                "name": "code",
                "language": "python",
                "options": (),
                "prompt_length": 3,
            },
            textwrap.dedent(
                """\
                .. code:: python

                   10 * 5
                """.rstrip()
            ),
            id="code",
        ),
        pytest.param(
            "10 * 5",
            {
                "name": "code-block",
                "language": "python",
                "options": (":okwarning:",),
                "prompt_length": 4,
            },
            textwrap.dedent(
                """\
                .. code-block:: python
                    :okwarning:

                    10 * 5
                """.rstrip()
            ),
            id="code-block_with_options",
        ),
        pytest.param(
            "10 * 5",
            {"name": "ipython", "language": None, "options": (), "prompt_length": 4},
            textwrap.dedent(
                """\
                .. ipython::

                    10 * 5
                """.rstrip()
            ),
            id="ipython",
        ),
    ),
)
def test_reformatting_func(code, directive, expected):
    actual = rst.reformatting_func(code, **directive)

    assert expected == actual


def test_blacken():
    labeled = tuple(
        ((min_ + 1, max_ + 1), label, "\n".join(data.lines[slice(min_, max_)]))
        for label, (min_, max_) in zip(data.line_labels, data.line_ranges)
    )
    actual = tuple(blacken(labeled))

    assert len("\n".join(actual).split("\n")) == 76



import pytest

from blackdoc import report


@pytest.mark.parametrize(
    ["sources", "expected"],
    (
        pytest.param(
            {
                "file1.py": "reformatted",
                "file2.py": "reformatted",
                "file3.rst": "reformatted",
            },
            (3, 0, 0),
            id="3 reformatted-0 unchanged-0 failed",
        ),
        pytest.param(
            {
                "file1.py": "unchanged",
                "file2.py": "unchanged",
                "file3.rst": "unchanged",
            },
            (0, 3, 0),
            id="0 reformatted-3 unchanged-0 failed",
        ),
        pytest.param(
            {"file1.py": "error", "file2.py": "error", "file3.rst": "error"},
            (0, 0, 3),
            id="0 reformatted-0 unchanged-3 failed",
        ),
        pytest.param(
            {"file1.py": "reformatted", "file2.py": "unchanged", "file3.rst": "error"},
            (1, 1, 1),
            id="1 reformatted-1 unchanged-1 failed",
        ),
    ),
)
def test_statistics(sources, expected):
    actual = report.statistics(sources)
    assert actual == expected


@pytest.mark.parametrize(
    ["statistics", "conditional", "expected"],
    (
        pytest.param(
            {"n_reformatted": 0, "n_unchanged": 1, "n_error": 0},
            False,
            ["1 file left unchanged"],
            id="0 reformatted-1 unchanged-0 failed",
        ),
        pytest.param(
            {"n_reformatted": 0, "n_unchanged": 1, "n_error": 0},
            True,
            ["1 file would be left unchanged"],
            id="0 reformatted-1 unchanged-0 failed-conditional",
        ),
        pytest.param(
            {"n_reformatted": 0, "n_unchanged": 4, "n_error": 0},
            False,
            ["4 files left unchanged"],
            id="0 reformatted-4 unchanged-0 failed",
        ),
        pytest.param(
            {"n_reformatted": 0, "n_unchanged": 4, "n_error": 0},
            True,
            ["4 files would be left unchanged"],
            id="0 reformatted-4 unchanged-0 failed-conditional",
        ),
        pytest.param(
            {"n_reformatted": 1, "n_unchanged": 0, "n_error": 0},
            False,
            ["1 file reformatted"],
            id="1 reformatted-0 unchanged-0 failed",
        ),
        pytest.param(
            {"n_reformatted": 1, "n_unchanged": 0, "n_error": 0},
            True,
            ["1 file would be reformatted"],
            id="1 reformatted-0 unchanged-0 failed-conditional",
        ),
        pytest.param(
            {"n_reformatted": 4, "n_unchanged": 0, "n_error": 0},
            False,
            ["4 files reformatted"],
            id="4 reformatted-0 unchanged-0 failed",
        ),
        pytest.param(
            {"n_reformatted": 4, "n_unchanged": 0, "n_error": 0},
            True,
            ["4 files would be reformatted"],
            id="4 reformatted-0 unchanged-0 failed-conditional",
        ),
        pytest.param(
            {"n_reformatted": 0, "n_unchanged": 0, "n_error": 1},
            False,
            ["1 file failed to reformat"],
            id="0 reformatted-0 unchanged-1 failed",
        ),
        pytest.param(
            {"n_reformatted": 0, "n_unchanged": 0, "n_error": 1},
            True,
            ["1 file would fail to reformat"],
            id="0 reformatted-0 unchanged-1 failed-conditional",
        ),
        pytest.param(
            {"n_reformatted": 0, "n_unchanged": 0, "n_error": 4},
            False,
            ["4 files failed to reformat"],
            id="0 reformatted-0 unchanged-4 failed",
        ),
        pytest.param(
            {"n_reformatted": 0, "n_unchanged": 0, "n_error": 4},
            True,
            ["4 files would fail to reformat"],
            id="0 reformatted-0 unchanged-4 failed-conditional",
        ),
        pytest.param(
            {"n_reformatted": 1, "n_unchanged": 1, "n_error": 1},
            False,
            [
                "1 file reformatted",
                "1 file left unchanged",
                "1 file failed to reformat",
            ],
            id="1 reformatted-1 unchanged-1 failed",
        ),
        pytest.param(
            {"n_reformatted": 1, "n_unchanged": 1, "n_error": 1},
            True,
            [
                "1 file would be reformatted",
                "1 file would be left unchanged",
                "1 file would fail to reformat",
            ],
            id="1 reformatted-1 unchanged-1 failed-conditional",
        ),
        pytest.param(
            {"n_reformatted": 4, "n_unchanged": 4, "n_error": 4},
            False,
            [
                "4 files reformatted",
                "4 files left unchanged",
                "4 files failed to reformat",
            ],
            id="4 reformatted-4 unchanged-4 failed",
        ),
        pytest.param(
            {"n_reformatted": 4, "n_unchanged": 4, "n_error": 4},
            True,
            [
                "4 files would be reformatted",
                "4 files would be left unchanged",
                "4 files would fail to reformat",
            ],
            id="4 reformatted-4 unchanged-4 failed-conditional",
        ),
    ),
)
def test_report_parts(statistics, conditional, expected):
    r = report.Report(conditional=conditional, **statistics)
    actual = r._report_parts()
    assert actual == expected



import textwrap

import more_itertools
import pytest

from blackdoc.formats import markdown


@pytest.mark.parametrize(
    ["string", "expected"],
    (
        pytest.param("", None, id="empty string"),
        pytest.param("Some string.", None, id="no_code"),
        pytest.param(
            textwrap.dedent(
                """\
                ```
                This is not a code block.
                ```
                """
            ),
            None,
            id="block",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ```sh
                find . -name "*.py"
                ```
                """
            ),
            None,
            id="code_other_language",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ```python
                >>> 10 * 5
                ```
                """
            ),
            None,
            id="doctest_prompt",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ```python
                In [1]: 10 * 5
                Out[1]:
                50
                ```
                """
            ),
            None,
            id="ipython_prompt",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ```python
                10 * 5
                ```
                """
            ),
            "markdown",
            id="code",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ``` python
                10 * 5
                ```
                """
            ),
            "markdown",
            id="code-space",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ```{python}
                10 * 5
                ```
                """
            ),
            "markdown",
            id="code-braces",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ```jupyter-execute
                10 * 5
                ```
                """
            ),
            "markdown",
            id="jupyter-execute",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ```{jupyter-execute}
                10 * 5
                ```
                """
            ),
            "markdown",
            id="jupyter-execute-braces",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ```{jupyter-execute}
                ---
                hide-code: true
                ---
                10 * 5
                ```
                """
            ),
            "markdown",
            id="jupyter-execute-braces-with_options",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ```{jupyter-execute}
                ---
                hide-code: true
                ---

                10 * 5
                ```
                """
            ),
            "markdown",
            id="jupyter-execute-braces-with_options-with_newlines",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                :::python
                10 * 5
                :::
                """
            ),
            "markdown",
            id="myst-code",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                :::{python}
                10 * 5
                :::
                """
            ),
            "markdown",
            id="myst-code-braces",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                :::{python}
                10 * 5
                :::
                """
            ),
            "markdown",
            id="myst-jupyter-execute",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                :::{python}
                10 * 5
                :::
                """
            ),
            "markdown",
            id="myst-jupyter-execute-braces",
        ),
    ),
)
def test_detection_func(string, expected):
    def construct_expected(label, string):
        if label is None:
            return None

        n_lines = len(string.split("\n"))

        range_ = (1, n_lines + 2)
        return range_, label, string

    lines = string.split("\n")
    code_fragment = more_itertools.peekable(enumerate(lines, start=1))
    actual = markdown.detection_func(code_fragment)

    assert actual == construct_expected(expected, string.rstrip())


@pytest.mark.parametrize(
    ["code", "expected"],
    (
        pytest.param(
            textwrap.dedent(
                """\
                ```python
                10 * 5
                ```
                """.rstrip()
            ),
            (
                {
                    "block_type": "python",
                    "prompt_length": 0,
                    "fences": "```",
                    "braces": False,
                    "options": (),
                },
                "10 * 5",
            ),
            id="backticks",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ``` python
                10 * 5
                ```
                """.rstrip()
            ),
            (
                {
                    "block_type": "python",
                    "prompt_length": 0,
                    "fences": "```",
                    "braces": False,
                    "options": (),
                },
                "10 * 5",
            ),
            id="backticks-with_space",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ```{python}
                10 * 5
                ```
                """.rstrip()
            ),
            (
                {
                    "block_type": "python",
                    "prompt_length": 0,
                    "fences": "```",
                    "braces": True,
                    "options": (),
                },
                "10 * 5",
            ),
            id="backticks-with_braces",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ```{jupyter-execute}
                10 * 5
                ```
                """.rstrip()
            ),
            (
                {
                    "block_type": "jupyter-execute",
                    "prompt_length": 0,
                    "fences": "```",
                    "braces": True,
                    "options": (),
                },
                "10 * 5",
            ),
            id="backticks-jupyter-execute-with_braces",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ```{jupyter-execute}
                ---
                hide-code: true
                hide-output: true
                ---
                10 * 5
                ```
                """.rstrip()
            ),
            (
                {
                    "block_type": "jupyter-execute",
                    "prompt_length": 0,
                    "fences": "```",
                    "braces": True,
                    "options": ("hide-code: true", "hide-output: true"),
                },
                "10 * 5",
            ),
            id="backticks-jupyter-execute-with_options",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                :::python
                10 * 5
                :::
                """.rstrip()
            ),
            (
                {
                    "block_type": "python",
                    "prompt_length": 0,
                    "fences": ":::",
                    "braces": False,
                    "options": (),
                },
                "10 * 5",
            ),
            id="colons",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ::: python
                10 * 5
                :::
                """.rstrip()
            ),
            (
                {
                    "block_type": "python",
                    "prompt_length": 0,
                    "fences": ":::",
                    "braces": False,
                    "options": (),
                },
                "10 * 5",
            ),
            id="colons-with_space",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                :::{python}
                10 * 5
                :::
                """.rstrip()
            ),
            (
                {
                    "block_type": "python",
                    "prompt_length": 0,
                    "fences": ":::",
                    "braces": True,
                    "options": (),
                },
                "10 * 5",
            ),
            id="colons-with_braces",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                :::{jupyter-execute}
                10 * 5
                :::
                """.rstrip()
            ),
            (
                {
                    "block_type": "jupyter-execute",
                    "prompt_length": 0,
                    "fences": ":::",
                    "braces": True,
                    "options": (),
                },
                "10 * 5",
            ),
            id="colons-jupyter-execute",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                :::python
                %load_ext extension
                10 * 5
                :::
                """.rstrip()
            ),
            (
                {
                    "block_type": "python",
                    "fences": ":::",
                    "braces": False,
                    "options": (),
                    "prompt_length": 0,
                },
                textwrap.dedent(
                    """\
                    # <ipython-magic>%load_ext extension
                    10 * 5
                    """.rstrip()
                ),
            ),
            id="ipython_magic",
        ),
    ),
)
def test_extraction_func(code, expected):
    actual = markdown.extraction_func(code)

    assert expected == actual


@pytest.mark.parametrize(
    ("code", "directive", "expected"),
    (
        pytest.param(
            "10 * 5",
            {
                "block_type": "python",
                "fences": "```",
                "options": (),
                "braces": False,
            },
            textwrap.dedent(
                """\
                ```python
                10 * 5
                ```
                """.rstrip()
            ),
            id="backticks",
        ),
        pytest.param(
            "10 * 5",
            {
                "block_type": "jupyter-execute",
                "fences": "```",
                "braces": True,
                "options": ("hide-code: true", "hide-output: true"),
            },
            textwrap.dedent(
                """\
                ```{jupyter-execute}
                ---
                hide-code: true
                hide-output: true
                ---
                10 * 5
                ```
                """.rstrip()
            ),
            id="jupyter-execute-with_options-with_braces",
        ),
        pytest.param(
            "10 * 5",
            {
                "block_type": "python",
                "fences": ":::",
                "braces": False,
                "options": (),
            },
            textwrap.dedent(
                """\
                :::python
                10 * 5
                :::
                """.rstrip()
            ),
            id="colons",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                # <ipython-magic>%load_ext extension
                10 * 5
                """.rstrip()
            ),
            {
                "block_type": "python",
                "fences": ":::",
                "braces": False,
                "options": (),
            },
            textwrap.dedent(
                """\
                :::python
                %load_ext extension
                10 * 5
                :::
                """.rstrip()
            ),
            id="ipython_magic",
        ),
    ),
)
def test_reformatting_func(code, directive, expected):
    actual = markdown.reformatting_func(code, **directive)

    assert expected == actual



def from_dict(labels):
    line_ranges = tuple(
        (
            (lineno - 1, lineno)
            if not isinstance(lineno, tuple)
            else tuple(n - 1 for n in lineno)
        )
        for lineno in labels.keys()
    )
    line_labels = tuple(labels.values())
    return line_ranges, line_labels


def to_classification_format(labels, lines):
    prepared_labels = dict(zip(*from_dict(labels)))
    return tuple(
        ((min_ + 1, max_ + 1), label, "\n".join(lines[min_:max_]))
        for (min_, max_), label in prepared_labels.items()
    )


def format_line_with_range(name, range_, unit):
    min_, max_ = range_
    line_numbers = range(min_, max_)

    no_group = " "
    start_group = "┐"
    mid_group = "│"
    end_group = "┘"

    lines = unit.split("\n")

    def determine_classifier(index):
        if max_ - min_ == 1:
            classifier = no_group
        elif index == 0:
            classifier = start_group
        elif index == max_ - min_ - 1:
            classifier = end_group
        else:
            classifier = mid_group

        return classifier

    return "\n".join(
        f"{name:>8s} {determine_classifier(index)} → {lineno:02d}: {line}"
        for index, (lineno, line) in enumerate(zip(line_numbers, lines))
    )


def format_classification(labeled):
    return "\n".join(
        format_line_with_range(range, name, unit) for name, range, unit in labeled
    )


def print_classification(labeled):
    print(format_classification(labeled))



import datetime
import difflib


def unified_diff(a, b, path):
    then = datetime.datetime.utcfromtimestamp(path.stat().st_mtime)
    now = datetime.datetime.utcnow()
    src_name = f"{path}\t{then} +0000"
    dst_name = f"{path}\t{now} +0000"

    diff = "\n".join(
        difflib.unified_diff(
            a.splitlines(),
            b.splitlines(),
            fromfile=src_name,
            tofile=dst_name,
            lineterm="",
        )
    )

    return diff



import textwrap

import more_itertools
import pytest

from blackdoc import blacken
from blackdoc.formats import doctest
from blackdoc.tests.data import doctest as data


@pytest.mark.parametrize(
    ("string", "expected"),
    (
        pytest.param("", None, id="empty string"),
        pytest.param("a", None, id="no quotes"),
        pytest.param("'''a'''", "'''", id="single quotes"),
        pytest.param('"""a"""', '"""', id="double quotes"),
        pytest.param('"a"""', None, id="trailing empty string"),
        pytest.param(
            textwrap.dedent(
                """\
                '''
                multiple lines
                '''
                """
            ).rstrip(),
            "'''",
            id="multiple lines single quotes",
        ),
        pytest.param(
            textwrap.dedent(
                '''\
                """
                multiple lines
                """
                '''
            ).rstrip(),
            '"""',
            id="multiple lines double quotes",
        ),
    ),
)
def test_detect_docstring_quotes(string, expected):
    actual = doctest.detect_docstring_quotes(string)
    assert expected == actual


@pytest.mark.parametrize(
    ["string", "expected"],
    (
        pytest.param("", None, id="empty_line"),
        pytest.param("a function to open lines", None, id="no_code"),
        pytest.param(">>>", "doctest", id="empty_line_with_prompt"),
        pytest.param(">>> 10 * 5", "doctest", id="single_line"),
        pytest.param("    >>> 10 * 5", "doctest", id="single_line_with_indent"),
        pytest.param(
            textwrap.dedent(
                """\
                >>> a = [
                ...     1,
                ...     2,
                ... ]
                """
            ),
            "doctest",
            id="multiple_lines",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                >>> ''' arbitrary triple-quoted string
                ...
                ... with empty continuation line
                ... '''
                """
            ),
            "doctest",
            id="multiple_lines_with_empty_continuation_line",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                >>>
                ... '''arbitrary triple-quoted string'''
                """
            ),
            "doctest",
            id="multiple_lines_with_leading_empty_continuation_line",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                >>> '''arbitrary triple-quoted string'''
                ...
                """
            ),
            "doctest",
            id="multiple_lines_with_trailing_empty_continuation_line",
        ),
    ),
)
def test_detection_func(string, expected):
    def construct_expected(label, string):
        if label is None:
            return None

        n_lines = len(string.split("\n"))
        range_ = (1, n_lines + 1)
        return range_, label, string

    lines = string.split("\n")
    lines_ = more_itertools.peekable(enumerate(lines, start=1))

    actual = doctest.detection_func(lines_)
    assert actual == construct_expected(expected, string.rstrip())


@pytest.mark.parametrize(
    ["code_unit", "expected"],
    (
        pytest.param(
            ">>>", ({"prompt_length": 4, "docstring_quotes": None}, ""), id="empty_line"
        ),
        pytest.param(
            ">>> 10 * 5",
            ({"prompt_length": 4, "docstring_quotes": None}, "10 * 5"),
            id="single_line",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                >>> a = [
                ...     1,
                ...     2,
                ...     3,
                ... ]
                """.rstrip()
            ),
            (
                {"prompt_length": 4, "docstring_quotes": None},
                textwrap.dedent(
                    """\
                    a = [
                        1,
                        2,
                        3,
                    ]
                    """.rstrip()
                ),
            ),
            id="multiple_lines",
        ),
        pytest.param(
            ">>> s = '''abc'''",
            (
                {"prompt_length": 4, "docstring_quotes": "'''"},
                "s = '''abc'''",
            ),
            id="triple_quoted_string-single_quotes",
        ),
        pytest.param(
            '>>> s = """abc"""',
            (
                {"prompt_length": 4, "docstring_quotes": '"""'},
                's = """abc"""',
            ),
            id="triple_quoted_string-double_quotes",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                >>> ''' arbitrary triple-quoted string
                ...
                ... with empty continuation line
                ... '''
                """.rstrip()
            ),
            (
                {"prompt_length": 4, "docstring_quotes": "'''"},
                textwrap.dedent(
                    """\
                    ''' arbitrary triple-quoted string

                    with empty continuation line
                    '''
                    """.rstrip()
                ),
            ),
            id="multiple_lines_with_empty_continuation_line",
        ),
    ),
)
def test_extraction_func(code_unit, expected):
    actual = doctest.extraction_func(code_unit)

    assert expected == actual


def prepare_lines(lines, remove_prompt=False):
    dedented = (line.lstrip() for line in more_itertools.always_iterable(lines))
    prepared = dedented if not remove_prompt else (line[4:] for line in dedented)
    return "\n".join(prepared)


@pytest.mark.parametrize(
    ["code_unit", "docstring_quotes", "expected"],
    (
        pytest.param(
            "file",
            None,
            ">>> file",
            id="single line",
        ),
        pytest.param(
            "",
            None,
            ">>>",
            id="single empty line",
        ),
        pytest.param(
            '"""docstring"""',
            '"""',
            '>>> """docstring"""',
            id="single-line triple-quoted string",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                a = [
                    1,
                    2,
                ]
                """.rstrip()
            ),
            None,
            textwrap.dedent(
                """\
                >>> a = [
                ...     1,
                ...     2,
                ... ]
                """.rstrip()
            ),
            id="multiple lines",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                print("abc")
                print("def")
                """
            ),
            None,
            textwrap.dedent(
                """\
                >>> print("abc")
                >>> print("def")
                """.rstrip()
            ),
            id="multiple lines multiple statements",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                # comment
                print("abc")
                """
            ),
            None,
            textwrap.dedent(
                """\
                >>> # comment
                ... print("abc")
                """.rstrip()
            ),
            id="multiple lines with comment",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                # comment
                def func():
                    pass
                """
            ),
            None,
            textwrap.dedent(
                """\
                >>> # comment
                ... def func():
                ...     pass
                ...
                """.rstrip()
            ),
            id="multiple lines comment before block",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                @decorator
                def func():
                    pass
                """
            ),
            None,
            textwrap.dedent(
                """\
                >>> @decorator
                ... def func():
                ...     pass
                ...
                """.rstrip()
            ),
            id="multiple lines function decorator",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                @decorator
                class A:
                    pass
                """
            ),
            None,
            textwrap.dedent(
                """\
                >>> @decorator
                ... class A:
                ...     pass
                ...
                """.rstrip()
            ),
            id="multiple lines class decorator",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                '''
                docstring content
                '''
                """.rstrip()
            ),
            "'''",
            textwrap.dedent(
                """\
                >>> '''
                ... docstring content
                ... '''
                """.rstrip()
            ),
            id="multi-line triple-quoted string-single quotes",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                s = '''
                    triple-quoted string
                '''
                """
            ).rstrip(),
            '"""',
            textwrap.dedent(
                '''\
                >>> s = """
                ...     triple-quoted string
                ... """
                '''.rstrip(),
            ),
            id="multi-line triple-quoted string-double quotes",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                ''' arbitrary triple-quoted string

                with a empty continuation line
                '''
                """.rstrip(),
            ),
            "'''",
            textwrap.dedent(
                """\
                >>> ''' arbitrary triple-quoted string
                ...
                ... with a empty continuation line
                ... '''
                """.rstrip(),
            ),
            id="multi-line triple-quoted string with empty continuation line",
        ),
        pytest.param(
            '"""inverted quotes"""',
            '"""',
            '>>> """inverted quotes"""',
            id="triple-quoted string with inverted quotes",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                def myfunc(arg1, arg2):
                    pass
                """
            ),
            None,
            textwrap.dedent(
                """\
                >>> def myfunc(arg1, arg2):
                ...     pass
                ...
                """.rstrip()
            ),
            id="trailing newline at the end of a block",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                a = 1

                """
            ),
            None,
            ">>> a = 1",
            id="trailing newline at the end of a normal line",
        ),
        pytest.param(
            "# this is not a block:",
            None,
            ">>> # this is not a block:",
            id="trailing colon at the end of a comment",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                def f(arg1, arg2):
                    ''' nested docstring

                    parameter description
                    '''
                """
            ),
            "'''",
            textwrap.dedent(
                """\
                >>> def f(arg1, arg2):
                ...     ''' nested docstring
                ...
                ...     parameter description
                ...     '''
                ...
                """.rstrip()
            ),
            id="nested docstring",
        ),
        pytest.param(
            "s = '''triple-quoted string'''",
            '"""',
            '>>> s = """triple-quoted string"""',
            id="triple-quoted string",
        ),
        pytest.param(
            textwrap.dedent(
                '''\
                def f(arg1, arg2):
                    """ docstring """
                    s = "trailing empty string"""
                '''
            ),
            "'''",
            textwrap.dedent(
                """\
                >>> def f(arg1, arg2):
                ...     ''' docstring '''
                ...     s = "trailing empty string\"""
                ...
                """.rstrip()
            ),
            id="docstring and trailing empty string",
        ),
        pytest.param(
            textwrap.dedent(
                '''\
                def f(arg1, arg2):
                    """ docstring """
                    s = """triple-quoted string"""
                '''
            ),
            "'''",
            textwrap.dedent(
                """\
                >>> def f(arg1, arg2):
                ...     ''' docstring '''
                ...     s = '''triple-quoted string'''
                ...
                """.rstrip()
            ),
            id="docstring and triple-quoted string",
        ),
    ),
)
def test_reformatting_func(code_unit, docstring_quotes, expected):
    actual = doctest.reformatting_func(code_unit, docstring_quotes)
    assert expected == actual

    # make sure the docstring quotes were not changed
    expected_quotes = doctest.detect_docstring_quotes(expected)
    actual_quotes = doctest.detect_docstring_quotes(actual)
    assert expected_quotes == actual_quotes


def test_blacken():
    labeled = tuple(
        ((min_ + 1, max_ + 1), label, "\n".join(data.lines[slice(min_, max_)]))
        for label, (min_, max_) in zip(data.line_labels, data.line_ranges)
    )
    actual = tuple(blacken(labeled))

    assert len("\n".join(actual).split("\n")) == 32



import itertools
import re
import textwrap

import more_itertools

from blackdoc.formats.errors import InvalidFormatError
from blackdoc.formats.ipython import hide_magic, reveal_magic
from blackdoc.formats.rst import has_prompt

name = "markdown"

# format:
# blocks followed by optional whitespace and the word python
# block fences are three backticks or colons (myst)
# the word can be wrapped by curly braces

directive_re = re.compile(
    r"""(?x)
    ^
    (?P<indent>[ ]*)
    (?P<fences>[`:]{3})
    \s*
    (?:
      (?P<braces>\{\s*(?P<block_type1>[-a-z0-9]+)\s*\})
      |(?P<block_type2>[-a-z0-9]+)
    )
    $
    """
)
include_pattern = r"\.md$"


def preprocess_directive(directive):
    block_type1 = directive.pop("block_type1")
    block_type2 = directive.pop("block_type2")

    new = dict(directive)
    new["block_type"] = block_type1 or block_type2
    new["braces"] = new["braces"] is not None

    return new


def take_while(iterable, predicate):
    while True:
        try:
            taken = next(iterable)
        except StopIteration:
            break

        if not predicate(taken):
            iterable.prepend(taken)
            break

        yield taken


def extract_options(lines, fences):
    taken = lines.peek()
    line = taken if isinstance(taken, str) else taken[1]

    if line.strip() != "---":
        return ()

    options = [next(lines)]
    # potentially found options
    while True:
        try:
            taken = next(lines)
        except StopIteration:
            break

        options.append(taken)

        line = taken if isinstance(taken, str) else taken[1]
        if line.strip() == "---":
            break
        elif line.strip() == fences:
            lines.prepend(*options)
            return ()

    return tuple(options)


def continuation_lines(lines, indent, fences):
    options = extract_options(lines, fences)
    newlines = tuple(take_while(lines, lambda x: not x[1].strip()))

    _, next_line = lines.peek((0, None))
    if has_prompt(next_line):
        lines.prepend(*options, *newlines)
        raise RuntimeError("prompt detected")

    yield from options
    yield from newlines
    yield from take_while(lines, lambda x: x[1].strip() != fences)


def detection_func(lines):
    try:
        line_number, line = lines.peek()
    except StopIteration:
        return None

    match = directive_re.match(line)
    if not match:
        return None

    directive = preprocess_directive(match.groupdict())
    if directive["block_type"] not in ("python", "python3", "jupyter-execute"):
        return None

    indent = len(directive.pop("indent"))

    start_line = more_itertools.first(lines)
    try:
        content_lines = list(continuation_lines(lines, indent, directive["fences"]))
    except RuntimeError as e:
        if str(e) != "prompt detected":
            raise

        lines.prepend((line_number, line))
        return None

    try:
        line_number, stop_line = lines.peek()
    except StopIteration:
        line_number = -1
        stop_line = None

    if stop_line.strip() != directive.get("fences"):
        raise RuntimeError("found a code block without closing fence")

    detected_lines = list(
        itertools.chain([start_line], content_lines, [more_itertools.first(lines)])
    )

    line_numbers, lines = map(tuple, more_itertools.unzip(detected_lines))
    line_range = min(line_numbers), max(line_numbers) + 2
    if line_numbers != tuple(range(line_range[0], line_range[1] - 1)):
        raise RuntimeError("line numbers are not contiguous")

    return line_range, name, "\n".join(lines)


def extraction_func(code):
    lines = more_itertools.peekable(iter(code.split("\n")))

    match = directive_re.fullmatch(more_itertools.first(lines))
    if not match:
        raise InvalidFormatError(f"misformatted code block:\n{code}")

    directive = preprocess_directive(match.groupdict())
    directive.pop("indent")
    directive["options"] = extract_options(lines, directive["fences"])[1:-1]

    lines_ = tuple(lines)
    if len(lines_) == 0:
        raise InvalidFormatError("misformatted code block: could not find any code")

    indent = len(lines_[0]) - len(lines_[0].lstrip())
    directive["prompt_length"] = indent

    code_ = textwrap.dedent("\n".join(lines_[:-1]))

    return directive, hide_magic(code_)


def reformatting_func(code, block_type, fences, braces, options):
    if braces:
        brace_open = "{"
        brace_close = "}"
    else:
        brace_open = ""
        brace_close = ""

    directive = f"{fences}{brace_open}{block_type}{brace_close}"
    parts = [directive]
    if options:
        parts.append(
            "\n".join(
                [
                    "---",
                    *options,
                    "---",
                ]
            )
        )
    parts.extend([reveal_magic(code), fences])

    return "\n".join(part for part in parts if part is not None)



import ast
import io
import itertools
import re
import sys
import tokenize
from tokenize import TokenError

import more_itertools

from blackdoc.formats.errors import InvalidFormatError

name = "doctest"
prompt_length = 4
prompt = ">>>"
prompt_re = re.compile(r"(>>> ?)")
continuation_prompt = "..."
continuation_prompt_re = re.compile(r"(\.\.\. ?)")
include_pattern = r"\.pyi?$"
block_start_re = re.compile(r"^[^#:]+:(\s*#.*)?$")


def continuation_lines(lines):
    while True:
        try:
            line_number, line = lines.peek()
        except StopIteration:
            line_number = -1
            line = ""

        if not continuation_prompt_re.match(line.lstrip()):
            break

        # actually consume the item
        more_itertools.consume(lines, n=1)

        yield line_number, line


def detection_func(lines):
    try:
        _, line = lines.peek()
    except StopIteration:
        line = ""

    if not prompt_re.match(line.lstrip()):
        return None

    detected_lines = list(
        itertools.chain([more_itertools.first(lines)], continuation_lines(lines))
    )
    line_numbers, lines = map(tuple, more_itertools.unzip(detected_lines))

    line_range = min(line_numbers), max(line_numbers) + 1
    if line_numbers != tuple(range(line_range[0], line_range[1])):
        raise InvalidFormatError("line numbers are not contiguous")

    return line_range, name, "\n".join(lines)


def suppress(iterable, errors):
    iter_ = iter(iterable)
    while True:
        try:
            yield next(iter_)
        except errors:
            yield None
        except StopIteration:
            break


def tokenize_string(code):
    readline = io.StringIO(code).readline

    return tokenize.generate_tokens(readline)


def extract_string_tokens(code):
    tokens = tokenize_string(code)

    # suppress invalid code errors: `black` will raise with a better error message
    return (
        token
        for token in suppress(tokens, TokenError)
        if token is not None and token.type == tokenize.STRING
    )


def detect_docstring_quotes(code_unit):
    def extract_quotes(string):
        if string.startswith("'''") and string.endswith("'''"):
            return "'''"
        elif string.startswith('"""') and string.endswith('"""'):
            return '"""'
        else:
            return None

    string_tokens = list(extract_string_tokens(code_unit))
    token_quotes = {token: extract_quotes(token.string) for token in string_tokens}
    quotes = (quote for quote in token_quotes.values() if quote is not None)

    return more_itertools.first(quotes, None)


def extraction_func(line):
    def extract_prompt(line):
        match = prompt_re.match(line)
        if match is not None:
            (prompt,) = match.groups()
            return prompt

        match = continuation_prompt_re.match(line)
        if match is not None:
            (prompt,) = match.groups()
            return prompt

        return ""

    def remove_prompt(line):
        prompt = extract_prompt(line)
        return line[len(prompt) :]

    lines = line.split("\n")
    if any(
        extract_prompt(line).rstrip() not in (prompt, continuation_prompt)
        for line in lines
    ):
        raise InvalidFormatError(f"misformatted code unit: {line}")

    extracted_line = "\n".join(remove_prompt(line) for line in lines)
    docstring_quotes = detect_docstring_quotes(extracted_line)

    return {
        "prompt_length": len(prompt) + 1,
        "docstring_quotes": docstring_quotes,
    }, extracted_line


def restore_quotes(code_unit, original_quotes):
    def line_offsets(code_unit):
        offsets = [m.end() for m in re.finditer("\n", code_unit)]

        return {lineno: offset for lineno, offset in enumerate([0] + offsets, start=1)}

    def compute_offset(pos, offsets):
        lineno, charno = pos
        return offsets[lineno] + charno

    if original_quotes is None:
        return code_unit

    to_replace = "'''" if original_quotes == '"""' else '"""'

    string_tokens = extract_string_tokens(code_unit)
    triple_quote_tokens = [
        token
        for token in string_tokens
        if token.string.startswith(to_replace) and token.string.endswith(to_replace)
    ]

    offsets = line_offsets(code_unit)
    mutable_string = io.StringIO(code_unit)
    for token in triple_quote_tokens:
        # find the offset in the stream
        start = compute_offset(token.start, offsets)
        end = compute_offset(token.end, offsets) - 3

        mutable_string.seek(start)
        mutable_string.write(original_quotes)

        mutable_string.seek(end)
        mutable_string.write(original_quotes)

    restored_code_unit = mutable_string.getvalue()

    return restored_code_unit


def split_by_statement(code_unit):
    """split a code unit into individual statements

    At this point, the only way to have more than a single statement
    is by joining multiple (non-block) statements with a `;`.
    """

    def lineno(node):
        # TODO: remove once we drop support for python=3.7
        version = (sys.version_info.major, sys.version_info.minor)

        if (
            version < (3, 8)
            and isinstance(node, ast.Expr)
            and isinstance(node.value, ast.Str)
        ):
            # bug in ast (fixed in py38): lineno is wrong for multi-line string expressions
            # https://bugs.python.org/issue16806
            n_lines = len(node.value.s.split("\n"))
            lineno = node.lineno - n_lines + 1
        elif isinstance(node, (ast.FunctionDef, ast.ClassDef)):
            linenos = [node.lineno] + [dec.lineno for dec in node.decorator_list]
            lineno = min(linenos)
        else:
            lineno = node.lineno

        return lineno

    content = ast.parse(code_unit).body

    lines = code_unit.split("\n")
    if not content:
        return [lines]

    indices = [lineno(obj) - 1 for obj in content]
    # make sure comments are included
    indices[0] = 0
    slices = more_itertools.zip_offset(indices, indices, offsets=(0, 1), longest=True)
    return [lines[start:stop] for start, stop in slices]


def reformatting_func(code_unit, docstring_quotes):
    def is_comment(line):
        return line.lstrip().startswith("#")

    def is_decorator(line):
        return line.lstrip().startswith("@")

    def drop_while(iterable, predicate):
        peekable = more_itertools.peekable(iterable)
        while True:
            try:
                current = peekable.peek()
            except StopIteration:
                break

            if not predicate(current):
                break

            more_itertools.consume(peekable, n=1)

        yield from peekable

    def is_block(lines):
        block_lines = drop_while(
            lines, lambda line: is_comment(line) or is_decorator(line)
        )
        first_line = more_itertools.first(block_lines, default="")
        match = block_start_re.match(first_line)
        return match is not None

    def add_prompt(prompt, line):
        if not line:
            return prompt

        return " ".join([prompt, line])

    def reformat_code_unit(lines):
        if is_block(lines):
            lines.append("")

        lines_ = iter(lines)
        return itertools.chain(
            more_itertools.always_iterable(
                add_prompt(prompt, more_itertools.first(lines_))
            ),
            (add_prompt(continuation_prompt, line) for line in lines_),
        )

    restored_quotes = restore_quotes(code_unit.rstrip(), docstring_quotes)

    subunits = split_by_statement(restored_quotes)

    return "\n".join(
        itertools.chain.from_iterable(reformat_code_unit(unit) for unit in subunits)
    )



import pytest

from blackdoc.blacken import parse_message


@pytest.mark.parametrize(
    "message,expected",
    (
        pytest.param(
            'Cannot parse: 16:10: with new_open("abc) as f:',
            ("Cannot parse", 16, 10, 'with new_open("abc) as f:'),
            id="simple_message",
        ),
    ),
)
def test_parse_message(message, expected):
    actual = parse_message(message)
    assert expected == actual



from black import Report

from blackdoc.blackcompat import (
    find_project_root,
    gen_python_files,
    get_gitignore,
    normalize_path_maybe_ignore,
)
from blackdoc.console import err


def collect_files(src, include, exclude, extend_exclude, force_exclude, quiet, verbose):
    root, _ = find_project_root(tuple(src))
    gitignore = get_gitignore(root)
    report = Report()

    for path in src:
        if path.is_dir():
            yield from gen_python_files(
                path.iterdir(),
                root,
                include,
                exclude,
                extend_exclude,
                force_exclude,
                report,
                gitignore,
                quiet=quiet,
                verbose=verbose,
            )
        elif str(path) == "-":
            yield path
        elif path.is_file():
            normalized_path = normalize_path_maybe_ignore(path, root, report)
            if normalized_path is None:
                continue

            normalized_path = "/" + normalized_path
            # Hard-exclude any files that matches the `--force-exclude` regex.
            if force_exclude:
                force_exclude_match = force_exclude.search(normalized_path)
            else:
                force_exclude_match = None
            if force_exclude_match and force_exclude_match.group(0):
                report.path_ignored(
                    path, "matches the --force-exclude regular expression"
                )
                continue

            yield path
        else:
            err.print(f"invalid path: {path}", style="red")



import pytest

from blackdoc import classification, formats
from blackdoc.tests import data
from blackdoc.tests.data import print_classification, to_classification_format


@pytest.mark.parametrize("format", ("rst", "doctest", "ipython"))
def test_detect_format(format, monkeypatch):
    module = getattr(formats, format)
    detection_funcs = {
        format: module.detection_func,
        "none": formats.none.detection_func,
    }
    monkeypatch.setattr(classification, "detection_funcs", detection_funcs)

    data_module = getattr(data, format)
    expected = to_classification_format(data_module.labels, data_module.lines)
    lines = enumerate(data_module.lines, start=1)
    actual = tuple(classification.detect_format(lines))

    print_classification(actual)

    assert expected == actual



from importlib.metadata import version

from blackdoc.blacken import blacken
from blackdoc.classification import detect_format
from blackdoc.formats import InvalidFormatError, register_format  # noqa: F401

try:
    __version__ = version("blackdoc")
except Exception:
    # Local copy or not installed with setuptools.
    # Disable minimum version checks on downstream libraries.
    __version__ = "999"


def line_numbers(lines):
    yield from enumerate(lines, start=1)


def format_lines(lines, mode=None):
    numbered = line_numbers(lines)

    labeled = detect_format(numbered)
    blackened = blacken(labeled, mode=mode)

    return blackened



import argparse
import importlib
import itertools
import sys

from blackdoc.tests.data.utils import format_classification


def format_conflicting_ranges(index_a, index_b, ranges):
    range_a = ranges[index_a]
    range_b = ranges[index_b]
    return f"{range_a} (item {index_a + 1}) ←→ {range_b} (item {index_b + 1})"


def display_data(lines, labels, ranges):
    labeled = tuple(
        ((start, stop), label, lines[start:stop])
        for (start, stop), label in zip(ranges, labels)
    )
    labeled_line_numbers = tuple(
        tuple(range(start, stop)) for (start, stop), _, _ in labeled
    )
    combinations = itertools.combinations(enumerate(labeled_line_numbers), 2)
    intersections = {
        (index_a, index_b): set(first).intersection(second)
        for (index_a, first), (index_b, second) in combinations
    }
    faulty_ranges = tuple(key for key, value in intersections.items() if value)
    if faulty_ranges:
        formatted_errors = [
            format_conflicting_ranges(index_a, index_b, ranges)
            for index_a, index_b in faulty_ranges
        ]
        sep = "\n -- "
        message = f"error: overlapping line ranges: {sep.join(formatted_errors)}"
        raise ValueError(message)

    covered_lines = tuple(itertools.chain.from_iterable(labeled_line_numbers))
    missing_lines = tuple(
        index for index, _ in enumerate(lines) if index not in covered_lines
    )
    missing_ranges = tuple((index, index + 1) for index in missing_lines)
    missing_label = "n/a"
    unlabeled = tuple(
        ((start, stop), missing_label, lines[start:stop])
        for start, stop in missing_ranges
    )
    combined = sorted(
        itertools.chain(labeled, unlabeled),
        key=lambda x: x[0][0],
    )
    return format_classification(
        tuple((range_, name, "\n".join(unit)) for range_, name, unit in combined)
    )


def display_module(module):
    exit_status = 0
    line_length = 80

    top_rule = "━" * line_length
    bottom_rule = "━" * line_length
    mid_rule = "─" * line_length

    try:
        formatted = display_data(module.lines, module.line_labels, module.line_ranges)
        print(top_rule)
        print(f"{{:^{line_length}}}".format("test data"))
        print(mid_rule)
        print(formatted)
        print(bottom_rule)
    except ValueError as e:
        print(e)
        exit_status += 1

    try:
        formatted = display_data(
            module.expected_lines,
            module.expected_line_labels,
            module.expected_line_ranges,
        )
        print("", "", sep="\n")
        print(top_rule)
        print(f"{{:^{line_length}s}}".format("expected output"))
        print(mid_rule)
        print(formatted)
        print(bottom_rule)
    except AttributeError:
        pass
    except ValueError as e:
        print(e)
        exit_status += 1

    return exit_status


parser = argparse.ArgumentParser()
parser.add_argument("format", help="print the data of this format")

args = parser.parse_args()
module = importlib.import_module(f".{args.format}", package="blackdoc.tests.data")
sys.exit(display_module(module))



import copy
import re

import black
from blib2to3.pgen2.tokenize import TokenError

from blackdoc.formats import extract_code, reformat_code


def parse_message(message):
    line_re = re.compile(
        r"^(?P<message>[^:]+): (?P<line_number>\d+):"
        r"(?P<column_number>\d+): (?P<faulty_line>.+)$"
    )

    types = {
        "message": str,
        "line_number": int,
        "column_number": int,
        "faulty_line": str,
    }

    match = line_re.match(message)
    if match is None:
        raise ValueError(f"invalid error message: {message}")

    return tuple(types[key](value) for key, value in match.groupdict().items())


def blacken(lines, mode=None):
    for original_line_range, code_format, line_unit in lines:
        if code_format == "none":
            yield line_unit
            continue

        indentation_depth, parameters, code = extract_code(line_unit, code_format)

        current_mode = black.FileMode() if mode is None else copy.copy(mode)
        current_mode.line_length -= indentation_depth + parameters.pop(
            "prompt_length", 0
        )

        original_line_number, _ = original_line_range
        original_line_number += parameters.pop("n_header_lines", 0)

        try:
            blackened = black.format_str(code, mode=current_mode).rstrip()
        except TokenError as e:
            message, (apparent_line_number, column) = e.args

            lineno = original_line_number + (apparent_line_number - 1)
            faulty_line = code.split("\n")[(apparent_line_number - 1) - 1]

            raise black.InvalidInput(
                f"Cannot parse: {lineno}:{column}: {message}: {faulty_line}"
            )
        except black.InvalidInput as e:
            message, apparent_line_number, column, faulty_line = parse_message(str(e))

            lineno = original_line_number + (apparent_line_number - 1)
            raise black.InvalidInput(f"{message}: {lineno}:{column}: {faulty_line}")
        except IndentationError as e:
            lineno = original_line_number + (e.lineno - 1)
            line = e.text.rstrip()

            # TODO: try to find the actual line, this exception is
            # only raised when the indentation causes the code to
            # become ambiguous
            raise black.InvalidInput(f"Invalid indentation: {lineno}: {line}")

        reformatted = reformat_code(
            blackened, code_format, indentation_depth, **parameters
        )

        yield reformatted



import textwrap

import more_itertools
import pytest

from blackdoc import blacken
from blackdoc.formats import ipython
from blackdoc.tests.data import ipython as data


@pytest.mark.parametrize(
    ["lines", "expected"],
    (
        pytest.param("xyz def", None, id="no_line"),
        pytest.param(
            "    In [2]: file",
            ((1, 2), ipython.name, "    In [2]: file"),
            id="single_line",
        ),
        pytest.param(
            [
                "In [1]: file = open(",
                '   ...:     "very_long_filepath",',
                '   ...:     mode="a",',
                "   ...: )",
            ],
            (
                (1, 5),
                ipython.name,
                textwrap.dedent(
                    """\
            In [1]: file = open(
               ...:     "very_long_filepath",
               ...:     mode="a",
               ...: )
            """.rstrip()
                ),
            ),
            id="multiple_lines",
        ),
    ),
)
def test_detection_func(lines, expected):
    lines = more_itertools.peekable(
        enumerate(more_itertools.always_iterable(lines), start=1)
    )

    actual = ipython.detection_func(lines)
    assert actual == expected


@pytest.mark.parametrize(
    ["line", "expected"],
    (
        pytest.param(
            "In [2]: file",
            ({"count": 2}, "file"),
            id="single_line",
        ),
        pytest.param(
            textwrap.dedent(
                """\
            In [1]: file = open(
               ...:     "very_long_filepath",
               ...:     mode="a",
               ...: )
            """.rstrip()
            ),
            (
                {"count": 1},
                "\n".join(
                    [
                        "file = open(",
                        '    "very_long_filepath",',
                        '    mode="a",',
                        ")",
                    ]
                ),
            ),
            id="multiple_lines",
        ),
        pytest.param(
            textwrap.dedent(
                """\
            In [4]: %%time
               ...: file.close()
            """.rstrip()
            ),
            (
                {"count": 4},
                textwrap.dedent(
                    f"""\
                # {ipython.magic_comment}%%time
                file.close()
                """.rstrip()
                ),
            ),
            id="lines_with_cell_magic",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                In [5]: @savefig simple.png width=4in
                   ...: @property
                   ...: def my_property(self):
                   ...:     pass
                """.rstrip()
            ),
            (
                {"count": 5},
                textwrap.dedent(
                    f"""\
                    # {ipython.magic_comment}@savefig simple.png width=4in
                    @property
                    def my_property(self):
                        pass
                    """.rstrip()
                ),
            ),
            id="lines_with_line_decorator",
        ),
    ),
)
def test_extraction_func(line, expected):
    actual = ipython.extraction_func(line)

    assert expected == actual


@pytest.mark.parametrize(
    ["line", "count", "expected"],
    (
        pytest.param("file", 2, "In [2]: file", id="single_line"),
        pytest.param(
            textwrap.dedent(
                """\
                file = open(
                    "very_long_filepath",
                    mode="a",
                )
                """.rstrip()
            ),
            1,
            textwrap.dedent(
                """\
                In [1]: file = open(
                   ...:     "very_long_filepath",
                   ...:     mode="a",
                   ...: )
                """.rstrip()
            ),
            id="multiple_lines",
        ),
        pytest.param(
            textwrap.dedent(
                f"""\
                # {ipython.magic_comment}%%time
                file.close()
                """.rstrip()
            ),
            4,
            textwrap.dedent(
                """\
                In [4]: %%time
                   ...: file.close()
                """.rstrip()
            ),
            id="lines_with_cell_magic",
        ),
        pytest.param(
            textwrap.dedent(
                f"""\
                # {ipython.magic_comment}@savefig simple.png width=4in
                @property
                def my_property(self):
                    pass
                """.rstrip()
            ),
            5,
            textwrap.dedent(
                """\
                In [5]: @savefig simple.png width=4in
                   ...: @property
                   ...: def my_property(self):
                   ...:     pass
                """.rstrip()
            ),
            id="lines_with_line_decorator",
        ),
    ),
)
def test_reformatting_func(line, count, expected):
    actual = ipython.reformatting_func(line, count=count)
    assert expected == actual


def test_blacken():
    labeled = list(
        (
            (min_, max_),
            label,
            "\n".join(data.lines[min_:max_]),
        )
        for (min_, max_), label in zip(data.line_ranges, data.line_labels)
    )
    actual = tuple(blacken(labeled))

    assert len(actual) == 19



import textwrap

import pytest
from rich.text import Span

from blackdoc import colors


@pytest.mark.parametrize(
    ["text", "spans"],
    (
        pytest.param(
            textwrap.dedent(
                """\
                - >>> a
                + >>> a + 1
                """.rstrip()
            ),
            [Span(0, 7, "red"), Span(8, 19, "green")],
            id="simple replacement",
        ),
        pytest.param(
            textwrap.dedent(
                f"""\
                - >>> a{' ' * 5}
                + >>> a + 1
                """.rstrip()
            ),
            [Span(0, 7, "red"), Span(7, 12, "red on red"), Span(13, 24, "green")],
            id="trailing whitespace",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                --- file1 time1
                +++ file2 time2
                """
            ),
            [Span(0, 15, "bold"), Span(16, 31, "bold")],
            id="header",
        ),
        pytest.param(
            "@@ line1,line2",
            [Span(0, 14, "cyan")],
            id="block header",
        ),
    ),
)
def test_diff_highlighter(text, spans):
    diff_highlighter = colors.DiffHighlighter()

    actual = diff_highlighter(text)
    assert actual.spans == spans


@pytest.mark.parametrize(
    ["text", "spans"],
    (
        pytest.param(
            "1 file would be reformatted",
            [Span(0, 6, "blue"), Span(0, 27, "bold")],
            id="single file-reformatted-conditional",
        ),
        pytest.param(
            "1 file reformatted",
            [Span(0, 6, "blue"), Span(0, 18, "bold")],
            id="single file-reformatted",
        ),
        pytest.param(
            "26 files would be reformatted",
            [Span(0, 8, "blue"), Span(0, 29, "bold")],
            id="multiple files-reformatted-conditional",
        ),
        pytest.param(
            "26 files reformatted",
            [Span(0, 8, "blue"), Span(0, 20, "bold")],
            id="multiple files-reformatted",
        ),
        pytest.param(
            "1 file would be left unchanged",
            [Span(0, 6, "blue")],
            id="single file-unchanged-conditional",
        ),
        pytest.param(
            "1 file left unchanged",
            [Span(0, 6, "blue")],
            id="single file-unchanged",
        ),
        pytest.param(
            "26 files would be left unchanged",
            [Span(0, 8, "blue")],
            id="multiple files-unchanged-conditional",
        ),
        pytest.param(
            "26 files left unchanged",
            [Span(0, 8, "blue")],
            id="multiple files-unchanged",
        ),
        pytest.param(
            "1 file would fail to reformat",
            [Span(0, 29, "red")],
            id="single file-error-conditional",
        ),
        pytest.param(
            "1 file failed to reformat",
            [Span(0, 25, "red")],
            id="single file-error",
        ),
        pytest.param(
            "15 files would fail to reformat",
            [Span(0, 31, "red")],
            id="multiple files-error-conditional",
        ),
        pytest.param(
            "15 files failed to reformat",
            [Span(0, 27, "red")],
            id="multiple files-error",
        ),
    ),
)
def test_file_highlighter(text, spans):
    highlighter = colors.FileHighlighter()

    actual = highlighter(text)
    assert actual.spans == spans



from blackdoc.tests.data.utils import from_dict

content = """\
Long description of the function's assumptions and on how to call it.

.. note::

    this is not a code block

As an example:

.. code:: python
   :okwarning:

   file = open(
       "very_long_filepath",
       mode="a"
   )


.. code-block:: python

    with open(
        "very_long_filepath",
        mode="a"
    ) as f:
        content = f.read()

A new example, this time with ipython:

.. ipython::

    %%time
    file = open(
        "very_long_filepath",
        mode="a"
    )
    file

with ipython prompts:

.. ipython::
    :okerror:

    In [1]: file = open(
       ...:     "very_long_filepath",
       ...:     mode="a"
       ...: )

    In [2]: file

    In [3]: file.read_binary()

with cell decorator:

.. ipython::
    :okerror:

    @verbatim
    In [1]: file = open(
       ...:     "very_long_filepath",
       ...:     mode="a"
       ...: )

    In [2]: file

    In [3]: file.read_binary()

a code block with a different language:

.. code:: sh

    find . -name "*.py"

a code block with testcode:

.. testsetup::

    file = open(
        "very_long_filepath",
        mode="a"
    )

.. testcode::

    file

.. testcleanup::

    file.close()
"""
lines = content.split("\n")
labels = {
    1: "none",
    2: "none",
    3: "none",
    4: "none",
    5: "none",
    6: "none",
    7: "none",
    8: "none",
    (9, 16): "rst",
    16: "none",
    17: "none",
    (18, 25): "rst",
    25: "none",
    26: "none",
    27: "none",
    (28, 36): "rst",
    36: "none",
    37: "none",
    38: "none",
    39: "none",
    40: "none",
    41: "none",
    42: "none",
    43: "none",
    44: "none",
    45: "none",
    46: "none",
    47: "none",
    48: "none",
    49: "none",
    50: "none",
    51: "none",
    52: "none",
    53: "none",
    54: "none",
    55: "none",
    56: "none",
    57: "none",
    58: "none",
    59: "none",
    60: "none",
    61: "none",
    62: "none",
    63: "none",
    64: "none",
    65: "none",
    66: "none",
    67: "none",
    68: "none",
    69: "none",
    70: "none",
    71: "none",
    72: "none",
    73: "none",
    (74, 80): "rst",
    80: "none",
    (81, 84): "rst",
    84: "none",
    (85, 88): "rst",
    88: "none",
}
line_ranges, line_labels = from_dict(labels)



import pytest

from blackdoc import autoupdate


@pytest.mark.parametrize(
    ["content", "expected"],
    (
        pytest.param(
            """\
            - repo: https://github.com/psf/black
              rev: 23.1.0
              hooks:
                - id: black
            """,
            "23.1.0",
        ),
        pytest.param(
            """\
            - repo: https://github.com/psf/black
              rev: 22.7.1
              hooks:
                - id: black-jupyter
            """,
            "22.7.1",
        ),
        pytest.param(
            """\
            - repo: https://github.com/psf/black-pre-commit-mirror
              rev: 23.10.1
              hooks:
                - id: black
            """,
            "23.10.1",
        ),
        pytest.param(
            """\
            - repo: https://github.com/psf/black-pre-commit-mirror
              rev: 22.9.10
              hooks:
                - id: black-jupyter
            """,
            "22.9.10",
        ),
        pytest.param(
            """\
            - repo: https://github.com/psf/black-pre-commit-mirror
              rev: 24.12.1
              hooks:
                - id: black
                - id: black-jupyter
            """,
            "24.12.1",
        ),
        pytest.param(
            """\
            - repo: https://github.com/psf/black-pre-commit-mirror
              rev: 24.12.1
              hooks:
                - id: black-jupyter
                - id: black
            """,
            "24.12.1",
        ),
    ),
)
def test_find_black_version(content, expected):
    version = autoupdate.find_black_version(content)

    assert version == expected


@pytest.mark.parametrize(
    ["content", "version", "expected"],
    (
        pytest.param(
            """\
            - repo: https://github.com/pre-commit/pre-commit-hooks
              rev: 4.4.0
              hooks:
                - id: trailing-whitespace
                - id: end-of-file-fixer

            - repo: https://github.com/keewis/blackdoc
              rev: 3.8.0
              hooks:
                - id: blackdoc
                  additional_dependencies: ["black==22.10.0"]
                - id: blackdoc-autoupdate-black
            """,
            "23.3.0",
            """\
            - repo: https://github.com/pre-commit/pre-commit-hooks
              rev: 4.4.0
              hooks:
                - id: trailing-whitespace
                - id: end-of-file-fixer

            - repo: https://github.com/keewis/blackdoc
              rev: 3.8.0
              hooks:
                - id: blackdoc
                  additional_dependencies: ["black==23.3.0"]
                - id: blackdoc-autoupdate-black
            """,
        ),
        pytest.param(
            """\
            - repo: https://github.com/keewis/blackdoc
              rev: 3.8.0
              hooks:
                - id: blackdoc
                  additional_dependencies: ["black==22.10.0"]
                - id: blackdoc-autoupdate-black
            """,
            "21.5.1",
            """\
            - repo: https://github.com/keewis/blackdoc
              rev: 3.8.0
              hooks:
                - id: blackdoc
                  additional_dependencies: ["black==21.5.1"]
                - id: blackdoc-autoupdate-black
            """,
        ),
        pytest.param(
            """\
            - repo: https://github.com/keewis/blackdoc
              rev: 3.8.0
              hooks:
                - id: blackdoc
                  additional_dependencies: ["black==22.10.0"]
                - id: blackdoc-autoupdate-black
            """,
            "23.7.12",
            """\
            - repo: https://github.com/keewis/blackdoc
              rev: 3.8.0
              hooks:
                - id: blackdoc
                  additional_dependencies: ["black==23.7.12"]
                - id: blackdoc-autoupdate-black
            """,
        ),
    ),
)
def test_update_black_pin(content, version, expected):
    updated = autoupdate.update_black_pin(content, version)

    assert updated == expected


@pytest.mark.parametrize(
    ["content", "expected"],
    (
        pytest.param(
            """\
        hooks:
        - repo: https://github.com/psf/black
          rev: 22.0.1
          hooks:
          - id: black

        - repo: https://github.com/keewis/blackdoc
          rev: 3.9.0
          hooks:
            - id: blackdoc
              additional_dependencies: ["black==22.0.1"]
        """,
            """\
        hooks:
        - repo: https://github.com/psf/black
          rev: 22.0.1
          hooks:
          - id: black

        - repo: https://github.com/keewis/blackdoc
          rev: 3.9.0
          hooks:
            - id: blackdoc
              additional_dependencies: ["black==22.0.1"]
        """,
        ),
        pytest.param(
            """\
        hooks:
        - repo: https://github.com/psf/black
          rev: 23.10.0
          hooks:
          - id: black

        - repo: https://github.com/keewis/blackdoc
          rev: 3.9.0
          hooks:
            - id: blackdoc
              additional_dependencies: ["black==22.0.1"]
        """,
            """\
        hooks:
        - repo: https://github.com/psf/black
          rev: 23.10.0
          hooks:
          - id: black

        - repo: https://github.com/keewis/blackdoc
          rev: 3.9.0
          hooks:
            - id: blackdoc
              additional_dependencies: ["black==23.10.0"]
        """,
        ),
    ),
)
def test_main(tmp_path, content, expected):
    path = tmp_path.joinpath(".pre-commit-config.yaml")
    path.write_text(content)

    stat = path.stat()

    return_value = autoupdate.main(path)

    updated = path.read_text()

    assert updated == expected
    if content == expected:
        assert path.stat() == stat
        assert return_value == 0
    else:
        assert return_value == 1



import argparse
import re

version_re = re.compile(
    r"https://github.com/psf/(?:black|black-pre-commit-mirror)\s+"
    r"rev: (.+)\s+hooks:(?:\s+-id: [-_a-zA-Z0-9]+)*\s+- id: (?:black|black-jupyter)"
)
black_pin_re = re.compile(
    r"(- id: blackdoc.+?additional_dependencies:.+?black==)[.\w]+",
    re.DOTALL,
)


def find_black_version(content):
    match = version_re.search(content)
    if match is None:
        raise ValueError("cannot find the black hook")
    version = match.group(1)
    return version


def update_black_pin(content, version):
    return black_pin_re.sub(rf"\g<1>{version}", content)


def main(path):
    with open(path) as f:
        content = f.read()

    version = find_black_version(content)
    replaced = update_black_pin(content, version)

    if content != replaced:
        with open(path, mode="w") as f:
            f.write(replaced)
        return 1
    else:
        return 0


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("path")
    args = parser.parse_args()
    raise SystemExit(main(args.path))



from blackdoc.tests.data.utils import from_dict

docstring = """ a function to open files

    with a very long description

    In [1]: file = open(
       ...:    "very_long_filepath",
       ...:    mode="a",
       ...: )

    In [2]: file
    Out[2]: <_io.TextIOWrapper name='very_long_filepath' mode='w' encoding='UTF-8'>

    text after the first example, spanning
    multiple lines

    In [3]: file.closed
    Out[3]: False

    In [4]: %%time
       ...: file.close()

    In [5]: @savefig simple.png width=4in
       ...: @property
       ...: def my_property(self):
       ...:     pass
"""
lines = docstring.split("\n")
labels = {
    1: "none",
    2: "none",
    3: "none",
    4: "none",
    (5, 9): "ipython",
    9: "none",
    10: "ipython",
    11: "none",
    12: "none",
    13: "none",
    14: "none",
    15: "none",
    16: "ipython",
    17: "none",
    18: "none",
    (19, 21): "ipython",
    21: "none",
    (22, 26): "ipython",
    26: "none",
}
line_ranges, line_labels = from_dict(labels)



from blackdoc.tests.data.utils import from_dict

docstring = """ a function to open files

    with a very long description

    >>> file = open(
    ...     "very_long_filepath",
    ...     mode="a",
    ... )
    >>> file
    <_io.TextIOWrapper name='very_long_filepath' mode='w' encoding='UTF-8'>

    text after the first example, spanning
    multiple lines

    >>> file.closed
    False

    >>> ''' arbitrary triple-quoted string
    ...
    ... with a empty continuation line
    ... '''
    >>> def myfunc2(arg1, arg2):
    ...     pass
    >>>

    >>> if myfunc2(2, 1) is not None:
    ...     print("caught")
    >>> a = 2
    ...
    >>> # this is not a block:
"""
lines = docstring.split("\n")
labels = {
    1: "none",
    2: "none",
    3: "none",
    4: "none",
    (5, 9): "doctest",
    9: "doctest",
    10: "none",
    11: "none",
    12: "none",
    13: "none",
    14: "none",
    15: "doctest",
    16: "none",
    17: "none",
    (18, 22): "doctest",
    (22, 24): "doctest",
    24: "doctest",
    25: "none",
    (26, 28): "doctest",
    (28, 30): "doctest",
    30: "doctest",
    31: "none",
}
line_ranges, line_labels = from_dict(labels)

expected = """ a function to open files

    with a very long description

    >>> file = open(
    ...     "very_long_filepath",
    ...     mode="a",
    ... )
    >>> file
    <_io.TextIOWrapper name='very_long_filepath' mode='w' encoding='UTF-8'>

    text after the first example, spanning
    multiple lines

    >>> file.closed
    False

    >>> ''' arbitrary triple-quoted string
    ...
    ... with a empty continuation line
    ... '''
    >>> def myfunc2(arg1, arg2):
    ...     pass
    ...
    >>>

    >>> if myfunc2(2, 1) is not None:
    ...     print("caught")
    ...
    >>> a = 2
    >>> # this is not a block:
"""
expected_lines = expected.split("\n")
expected_labels = {
    1: "none",
    2: "none",
    3: "none",
    4: "none",
    (5, 9): "doctest",
    9: "doctest",
    10: "none",
    11: "none",
    12: "none",
    13: "none",
    14: "none",
    15: "doctest",
    16: "none",
    17: "none",
    (18, 22): "doctest",
    (22, 25): "doctest",
    25: "doctest",
    26: "none",
    (27, 30): "doctest",
    30: "doctest",
    31: "doctest",
    32: "none",
}
expected_line_ranges, expected_line_labels = from_dict(expected_labels)



import textwrap

import more_itertools

from blackdoc.formats import doctest, ipython, markdown, none, rst
from blackdoc.formats.errors import InvalidFormatError  # noqa: F401
from blackdoc.formats.register import (
    detection_funcs,  # noqa: F401
    disable,  # noqa: F401
    extraction_funcs,
    format_include_patterns,  # noqa: F401
    include_patterns,  # noqa: F401
    reformatting_funcs,
    register_format,
)


def extract_code(line_unit, code_format):
    dedented = textwrap.dedent(line_unit)
    indentation_depth = len(more_itertools.first(line_unit.split("\n"))) - len(
        more_itertools.first(dedented.split("\n"))
    )

    func = extraction_funcs.get(code_format, None)
    if func is None:
        raise RuntimeError(f"unknown code format: {code_format}")

    parameters, extracted = func(dedented)
    return indentation_depth, parameters, extracted


def reformat_code(line_unit, code_format, indentation_depth, **parameters):
    func = reformatting_funcs.get(code_format, None)
    if func is None:
        raise RuntimeError(f"unknown code format: {code_format}")

    reformatted = func(line_unit, **parameters)

    return textwrap.indent(reformatted, " " * indentation_depth)


for module in (none, doctest, ipython, rst, markdown):
    register_format(module.name, module)



# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------
import datetime as dt

project = "blackdoc"
author = f"{project} developers"
year = dt.datetime.now().year
copyright = f"2020-{year}, {author}"

# The master toctree document.
master_doc = "index"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
]

extlinks = {
    "issue": ("https://github.com/keewis/blackdoc/issues/%s", "GH%s"),
    "pull": ("https://github.com/keewis/blackdoc/pull/%s", "PR%s"),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "directory"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# -- Options for the intersphinx extension -----------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/stable/", None),
}



from blackdoc.formats import none
from blackdoc.tests.data.doctest import lines


def test_detection_func():
    line_range = (1, 2)
    line = lines[0]
    name = none.name

    assert none.detection_func(enumerate(lines, start=1)) == (line_range, name, line)


def test_extraction_func():
    parameters = {}
    line = lines[0]

    assert none.extraction_func(line) == (parameters, line)


def test_reformatting_func():
    line = lines[0]

    assert none.reformatting_func(line) == line



import more_itertools

from blackdoc.formats import detection_funcs


def detect_format(lines):
    lines = more_itertools.peekable(lines)
    while lines:
        maybe_detected = (
            (name, func(lines))
            for name, func in detection_funcs.items()
            if name != "none"
        )
        detected = {name: value for name, value in maybe_detected if value is not None}

        if not detected:
            yield detection_funcs["none"](lines)
        elif len(detected) > 1:
            formatted_match_names = ", ".join(sorted(detected.keys()))
            raise RuntimeError(
                "cannot detect code format for line:"
                f" it is claimed by {formatted_match_names}: {lines.peek()}"
            )
        else:
            yield more_itertools.one(detected.values())



import itertools
import re
import textwrap

import more_itertools

from blackdoc.formats.doctest import prompt_re as doctest_prompt_re
from blackdoc.formats.errors import InvalidFormatError
from blackdoc.formats.ipython import hide_magic, reveal_magic
from blackdoc.formats.ipython import prompt_re as ipython_prompt_re

name = "rst"

directive_re = re.compile(
    "(?P<indent>[ ]*).. (?P<name>[a-z][-a-z]*)::(?: (?P<language>[a-z]+))?"
)
option_re = re.compile(r"^\s*:[^:]+:")

include_pattern = r"\.rst$"


def has_prompt(line):
    return any(
        prompt_re.match(line.lstrip())
        for prompt_re in [ipython_prompt_re, doctest_prompt_re]
    )


def take_while(iterable, predicate):
    while True:
        try:
            taken = next(iterable)
        except StopIteration:
            break

        if not predicate(taken):
            iterable.prepend(taken)
            break

        yield taken


def continuation_lines(lines, indent):
    options = tuple(take_while(lines, lambda x: option_re.match(x[1])))
    newlines = tuple(take_while(lines, lambda x: not x[1].strip()))
    decorator_lines = tuple(take_while(lines, lambda x: x[1].lstrip().startswith("@")))
    _, next_line = lines.peek((0, None))
    if next_line is None:
        return

    if has_prompt(next_line):
        lines.prepend(*options, *newlines, *decorator_lines)
        raise RuntimeError("prompt detected")

    yield from options
    yield from newlines
    yield from decorator_lines

    while True:
        newlines = tuple(take_while(lines, lambda x: not x[1].strip()))
        try:
            line_number, line = lines.peek()
        except StopIteration:
            lines.prepend(*newlines)
            break

        current_indent = len(line) - len(line.lstrip())
        if current_indent <= indent:
            # put back the newlines, if any
            lines.prepend(*newlines)
            break

        yield from newlines

        # consume the line
        more_itertools.consume(lines, n=1)

        yield line_number, line


def detection_func(lines):
    try:
        line_number, line = lines.peek()
    except StopIteration:
        return None

    match = directive_re.match(line)
    if not match:
        return None

    directive = match.groupdict()

    if directive["name"] not in (
        "code",
        "code-block",
        "ipython",
        "testcode",
        "testsetup",
        "testcleanup",
    ):
        return None

    if directive["language"] not in ("python", None):
        return None

    indent = len(directive.pop("indent"))

    try:
        detected_lines = list(
            itertools.chain(
                [more_itertools.first(lines)],
                continuation_lines(lines, indent),
            )
        )
    except RuntimeError as e:
        if str(e) != "prompt detected":
            raise

        lines.prepend((line_number, line))
        return None

    line_numbers, lines = map(tuple, more_itertools.unzip(detected_lines))
    line_range = min(line_numbers), max(line_numbers) + 1
    if line_numbers != tuple(range(line_range[0], line_range[1])):
        raise RuntimeError("line numbers are not contiguous")

    return line_range, name, "\n".join(lines)


def extraction_func(code):
    lines = more_itertools.peekable(iter(code.split("\n")))

    match = directive_re.fullmatch(more_itertools.first(lines))
    if not match:
        raise InvalidFormatError(f"misformatted code block:\n{code}")

    directive = match.groupdict()
    directive.pop("indent")

    directive["options"] = tuple(
        line.strip() for line in take_while(lines, lambda line: option_re.match(line))
    )

    # correct a missing newline
    newline = lines.peek(None)
    if newline is None:
        raise InvalidFormatError(
            "misformatted code block:"
            " newline after directive options required"
            " but found <end-of-file>"
        )
    elif not newline.strip():
        more_itertools.first(lines)

    lines_ = tuple(lines)
    if len(lines_) == 0:
        raise InvalidFormatError("misformatted code block: could not find any code")

    indent = len(lines_[0]) - len(lines_[0].lstrip())
    directive["prompt_length"] = indent
    directive["n_header_lines"] = len(directive["options"]) + 2

    code_ = hide_magic(textwrap.dedent("\n".join(lines_)))

    return directive, code_


def reformatting_func(code, name, language, options, prompt_length=4):
    indent = " " * prompt_length

    directive = " ".join(
        [f".. {name}::"] + ([language] if language is not None else [])
    )

    options_ = textwrap.indent("\n".join(options), indent) if options else None
    code_ = textwrap.indent(reveal_magic(code), indent)

    return "\n".join(
        line for line in (directive, options_, "", code_) if line is not None
    )



from rich.text import Text

from blackdoc.colors import FileHighlighter

highlighter = FileHighlighter()


def statistics(sources):
    from collections import Counter

    statistics = Counter(sources.values())

    n_unchanged = statistics.pop("unchanged", 0)
    n_reformatted = statistics.pop("reformatted", 0)
    n_error = statistics.pop("error", 0)

    if len(statistics) != 0:
        raise RuntimeError(f"unknown results: {statistics.keys()}")

    return n_reformatted, n_unchanged, n_error


def noun(n):
    return "file" if n < 2 else "files"


def _report_words(report_type, conditional):
    mapping = {
        "reformatted": {
            False: "reformatted",
            True: "would be reformatted",
        },
        "unchanged": {
            False: "left unchanged",
            True: "would be left unchanged",
        },
        "error": {
            False: "failed to reformat",
            True: "would fail to reformat",
        },
    }

    return mapping.get(report_type, {}).get(conditional)


class Report:
    def __init__(self, n_reformatted, n_unchanged, n_error, conditional=False):
        self.n_reformatted = n_reformatted
        self.n_unchanged = n_unchanged
        self.n_error = n_error

        self.conditional = conditional

    @classmethod
    def from_sources(cls, sources, conditional=False):
        n_reformatted, n_unchanged, n_error = statistics(sources)

        return cls(n_reformatted, n_unchanged, n_error, conditional=conditional)

    def __repr__(self):
        params = [
            f"{name}={getattr(self, name)}"
            for name in ["n_reformatted", "n_unchanged", "n_error", "conditional"]
        ]
        return f"Report({', '.join(params)})"

    def _report_parts(self):
        report_types = ["reformatted", "unchanged", "error"]
        values = {
            report_type: getattr(self, f"n_{report_type}")
            for report_type in report_types
        }
        parts = [
            f"{value} {noun(value)} {_report_words(report_type, self.conditional)}"
            for report_type, value in values.items()
            if value > 0
        ]
        return parts

    def __str__(self):
        parts = self._report_parts()
        return ", ".join(parts) + "."

    def __rich__(self):
        parts = [highlighter(part) for part in self._report_parts()]
        return Text(", ").join(parts) + Text(".")



import warnings

import more_itertools

detection_funcs = {}
extraction_funcs = {}
reformatting_funcs = {}
include_patterns = {}


def format_include_patterns():
    patterns = set(include_patterns.values())
    joined_patterns = "|".join(patterns)

    if "|" not in joined_patterns:
        return joined_patterns
    else:
        return f"({joined_patterns})"


def disable(format_names):
    names = tuple(more_itertools.always_iterable(format_names))
    unknown_names = tuple(name for name in names if name not in detection_funcs)
    if any(unknown_names):
        raise ValueError(f"unknown formats: {','.join(unknown_names)}")

    for name in names:
        del detection_funcs[name]


def register_format(name, obj):
    """register a new format"""
    if name in detection_funcs:
        warnings.warn(f"{name} already registered", RuntimeWarning)

    detection_func = getattr(obj, "detection_func")
    extraction_func = getattr(obj, "extraction_func")
    reformatting_func = getattr(obj, "reformatting_func")
    include_pattern = getattr(obj, "include_pattern", None)

    detection_funcs[name] = detection_func
    extraction_funcs[name] = extraction_func
    reformatting_funcs[name] = reformatting_func

    if include_pattern is not None:
        include_patterns[name] = include_pattern



import itertools
import re

from rich.highlighter import Highlighter

line_re = re.compile("\n")
trailing_whitespace_re = re.compile(r"\s+$")


def line_style(lineno, line):
    if line.startswith("+++") or line.startswith("---"):
        yield lineno, (0, len(line)), "bold"
    elif line.startswith("@@"):
        yield lineno, (0, len(line)), "cyan"
    elif line.startswith("+"):
        yield lineno, (0, len(line)), "green"
    elif line.startswith("-"):
        yield lineno, (0, len(line.rstrip())), "red"
        trailing_whitespace = trailing_whitespace_re.search(line)
        if trailing_whitespace:
            start, end = trailing_whitespace.span()
            yield lineno, (start, end), "red on red"
    else:
        yield lineno, (0, len(line)), ""


def line_offsets(text):
    matches = line_re.finditer(text)

    return [0] + [m.end() for m in matches]


def move_span(start, end, offset):
    return start + offset, end + offset


class DiffHighlighter(Highlighter):
    def highlight(self, text):
        def diff_styles(text):
            lines = text.split("\n")
            line_styles = itertools.chain.from_iterable(
                line_style(lineno, line) for lineno, line in enumerate(lines, start=1)
            )

            offsets = line_offsets(text)
            styles = {
                move_span(start, end, offsets[lineno - 1]): style
                for lineno, (start, end), style in line_styles
            }

            yield from styles.items()

        for (start, end), style in diff_styles(text.plain):
            text.stylize(style, start=start, end=end)


class FileHighlighter(Highlighter):
    highlights = {
        r"[0-9]+ files?(?!.*fail)": "blue",
        r"^.+reformatted$": "bold",
        r"^.+fail.+$": "red",
    }

    def highlight(self, text):
        for highlight_re, style in self.highlights.items():
            text.highlight_regex(highlight_re, style=style)



import more_itertools

name = "none"


def detection_func(lines):
    number, line = more_itertools.first(lines)
    return (number, number + 1), name, line


def extraction_func(line):
    return {}, line


def reformatting_func(line):
    return line
